use chrono_tz::Tz;
use poise::serenity_prelude::{model::id::GuildId, ResolvedValue};
use serde::{Deserialize, Serialize};
use serde_json::Value;

use crate::{
    commands::remind::RemindOptions, models::reminder::create_reminder, ApplicationContext,
    Context, Error,
};

#[derive(Serialize, Deserialize)]
#[serde(tag = "command_name")]
pub enum RecordedCommand {
    Remind(RemindOptions),
}

macro_rules! extract_arg {
    ($ctx:ident, $name:literal, String) => {
        $ctx.args.iter().find(|opt| opt.name == $name).map(|opt| &opt.value).map_or_else(
            || String::new(),
            |v| match v {
                ResolvedValue::String(s) => s.to_string(),
                _ => String::new(),
            },
        )
    };
    ($ctx:ident, $name:literal, Option<String>) => {
        $ctx.args.iter().find(|opt| opt.name == $name).map(|opt| &opt.value).map(|v| match v {
            ResolvedValue::String(s) => s.to_string(),
            _ => String::new(),
        })
    };
    ($ctx:ident, $name:literal, bool) => {
        $ctx.args.iter().find(|opt| opt.name == $name).map(|opt| &opt.value).map(|v| match v {
            ResolvedValue::Boolean(b) => b.to_owned(),
            _ => false,
        })
    };
    ($ctx:ident, $name:literal, Option<Tz>) => {
        $ctx.args
            .iter()
            .find(|opt| opt.name == $name)
            .map(|opt| &opt.value)
            .map(|v| match v {
                ResolvedValue::String(s) => s.parse::<Tz>().ok(),
                _ => None,
            })
            .flatten()
    };
}

impl RecordedCommand {
    pub fn from_context(ctx: ApplicationContext) -> Option<Self> {
        match ctx.command().identifying_name.as_str() {
            "remind" => Some(Self::Remind(RemindOptions {
                time: extract_arg!(ctx, "time", String),
                content: extract_arg!(ctx, "content", String),
                channels: extract_arg!(ctx, "channels", Option<String>),
                interval: extract_arg!(ctx, "interval", Option<String>),
                expires: extract_arg!(ctx, "expires", Option<String>),
                tts: extract_arg!(ctx, "tts", bool),
                timezone: extract_arg!(ctx, "timezone", Option<Tz>),
            })),
            _ => None,
        }
    }

    pub async fn execute(self, ctx: ApplicationContext<'_>) -> Result<(), Error> {
        match self {
            RecordedCommand::Remind(command_options) => {
                create_reminder(
                    Context::Application(ctx),
                    command_options.time,
                    command_options.content,
                    command_options.channels,
                    command_options.interval,
                    command_options.expires,
                    command_options.tts,
                    command_options.timezone,
                )
                .await
            }
        }
    }
}

pub struct CommandMacro {
    pub guild_id: GuildId,
    pub name: String,
    pub description: Option<String>,
    pub commands: Vec<RecordedCommand>,
}

pub async fn guild_command_macro(ctx: &Context<'_>, name: &str) -> Option<CommandMacro> {
    struct Row {
        name: String,
        description: Option<String>,
        commands: Value,
    }

    let row = sqlx::query_as!(
        Row,
        "
        SELECT m.name, m.description, m.commands
        FROM command_macro m
        INNER JOIN guilds g
            ON g.id = m.guild_id
        WHERE guild = ?
            AND m.name = ?
        ",
        ctx.guild_id().unwrap().get(),
        name
    )
    .fetch_one(&ctx.data().database)
    .await
    .ok()?;

    let command_macro = CommandMacro {
        guild_id: ctx.guild_id().unwrap(),
        name: row.name,
        description: row.description,
        commands: serde_json::from_value(row.commands).unwrap(),
    };

    Some(command_macro)
}
